"""
Provides bgameb version information.
"""

# This file is auto-generated! Do not edit!
# Use `python -m incremental.update bgameb` to change this file.

from incremental import Version

__version__ = Version("bgameb", 2, 0, 0)
__all__ = ["__version__"]
